#include "Scratchpad.h"

#include "../pointers/pointers.h"
#include <stdlib.h>
#include <string.h>

int Scratchpad_Create(Scratchpad* target, size_t capacity, allocator_t* allocator)
{
	if (target == NULL) return EDESTADDRREQ;

	void* memory = Allocator_Allocate(allocator, capacity);
	if (memory == NULL) return ENOMEM;

	target->memory = memory;
	target->capacity = capacity;
	target->reserved = 0;
	target->next     = NULL;
	target->allocator = allocator;

	return EXIT_SUCCESS;
}

static void* _Scratchpad_ReserveHere(Scratchpad* pad, size_t size)
{
	if (size > pad->capacity - pad->reserved) return NULL;

	void* scratch_buffer = advancep(pad->memory, pad->reserved);
	pad->reserved += size;

	return scratch_buffer;
}

void* Scratchpad_Reserve(Scratchpad* pad, size_t size)
{
	if (pad == NULL) return NULL;

	Scratchpad* current = pad;
	void* reserve = NULL;
	while (reserve == NULL && current != NULL) {
		reserve = _Scratchpad_ReserveHere(current, size);
		current = current->next;
	}

	if (reserve == NULL) {
		Scratchpad* new = Allocator_Allocate(pad->allocator, sizeof(*pad));
		if (new != NULL) {
			// Copy current pad in the *next pointer
			memcpy(new, pad, sizeof(*pad));

			// Calculate new capacity
			size_t new_capacity = size > pad->capacity ?
				pad->capacity * (size / pad->capacity + 1)
				: pad->capacity;

			// Initialize new pad
			if (Scratchpad_Create(pad, new_capacity, pad->allocator)) {
				Allocator_Free(pad->allocator, new, sizeof(*new));
			} else {
				reserve = Scratchpad_Reserve(pad, size);
				pad->next = new;
			}
		}
	}

	return reserve;
}

int Scratchpad_Reclaim(Scratchpad* pad, void* pointer, size_t length)
{
	if (pad == NULL) return EDESTADDRREQ;
	if (pointer == NULL) return EINVAL;
	if (length > pad->reserved) return EBOUNDS;

	if (rewindp(advancep(pad->memory, pad->reserved), length) == pointer) {
		pad->reserved -= length;
	}

	return EXIT_SUCCESS;
}

static void _Scratchpad_DestroyHere(Scratchpad* pad)
{
	Allocator_Free(pad->allocator, pad->memory, pad->capacity);

	return;
}

void Scratchpad_Reset(Scratchpad* pad)
{
	if (pad == NULL) return;

	while (pad->next != NULL && pad->next->next != NULL) {
		Scratchpad* current = pad->next;
		pad->next = current->next;

		_Scratchpad_DestroyHere(current);
		Allocator_Free(pad->allocator, current, sizeof(*current));
	}

	if (pad->next != NULL) {
		Scratchpad* next = pad->next;
		_Scratchpad_DestroyHere(pad);
		memcpy(pad, next, sizeof(*next));

		Allocator_Free(pad->allocator, next, sizeof(*next));
	}

	return;
}

void Scratchpad_Destroy(Scratchpad* pad)
{
	if (pad == NULL) return;

	while (pad->next != NULL) {
		Scratchpad* next_next = pad->next->next;
		_Scratchpad_DestroyHere(pad->next);
		Allocator_Free(pad->allocator, pad->next, sizeof(*pad));

		pad->next = next_next;
	}

	_Scratchpad_DestroyHere(pad);

	return;
}
