#ifndef V_THREADPOOL_H
#define V_THREADPOOL_H

#include <stdint.h> // int32_t, size_t, int8_t
#include <stdbool.h> // bool

#include "../allocator-interface/allocator-interface.h" // for allocator
#include "../FreeList/FreeList.h"

#include "../threading/os_semaphore.h" // for os_semaphore_t
#include "../threading/os_thread.h" // for os_thread_t
#include "../threading/os_mutex.h" // for os_mutex_t

#define THREADPOOLJOB_DISCARD_RESULT (1)

typedef struct ThreadPoolJob_s {
	ThreadFunction job;
	void* result;
	void* arg;
	int8_t flags;
	size_t next;
} ThreadPoolJob;

typedef struct ThreadPoolWorker_s {
	bool 		alive;
	os_thread_t thread;
	size_t 		job;
} ThreadPoolWorker;

#define THREADPOOL_KILL_WORKERS (1)

typedef struct ThreadPool_s {
	os_mutex_t     rw_lock;
	os_semaphore_t wait_queue;

	ThreadPoolWorker*	workers;
	size_t			worker_count;

	FreeList       job_storage;

	size_t first_job;
	size_t last_job;

	size_t first_result;
	size_t last_result;

	int8_t        flags;

	allocator_t*   allocator;
} ThreadPool;

/*!
   @brief Create a ThreadPool and spin up the workers, unless KILL_WORKERS-flag is given
   @param target this
   @param worker_count How many worker threads should the thread pool use
   @param flags Flag integer (use OR conjunction)
   @param allocator for the thread-pool
   @return EXIT_SUCCESS, ENOMEM
*/
int ThreadPool_Create(ThreadPool* target, size_t worker_count, int8_t flags, allocator_t* allocator);

/*!
   @brief Queue a Job for completion in the ThreadPool
   @param pool this
   @param job The Job to queue, will be copied into ThreadPool Memory
   @param job_id_out (out) The assigned job_id will be placed here
   @return EXIT_SUCCESS, ENOMEM
*/
int ThreadPool_QueueJob(ThreadPool* pool, ThreadPoolJob* job, size_t* job_id_out);


/*!
   @brief Remove a queued job from the queue
   @param pool this
   @param job_id job_id of the job to cancel
   @return EXIT_SUCCESS, EINVAL, EINPROGRESS
*/
int ThreadPool_UnqueueJob(ThreadPool* pool, size_t job_id);

/*!
   @brief Query whether a job has finished already
   @param pool this
   @param job_id job_id of the queried job result
   @return EXIT_SUCCESS, EINVAL, EINPROGRESS
*/
bool ThreadPool_HasFinished(ThreadPool* pool, size_t job_id);

/*!
   @brief Query whether a job has finished already
   @param pool this
   @param job_id job_id of the queried job result
   @param job_result (out) job result will be placed here
   @pre the associated job must have finished
   @post the job result may not be queried again
   @return EXIT_SUCCESS, EINVAL, EINPROGRESS
*/
int ThreadPool_GetJobResult(ThreadPool* pool, size_t job_id, void** job_result);

/*!
   @brief Retrieve the ID of the result finished for the longest time
   @param pool this
   @pre pool is not NULL
   @return size_t, SIZE_MAX (no result)
*/
size_t ThreadPool_GetNextResultID(ThreadPool* pool);

/*!
   @brief Destroy the thread-pool and all associated resources
   @param thread_pool this
   @return void
*/
void ThreadPool_Destroy(ThreadPool* thread_pool);

#endif
