/*
 * This code is part of the strategy game operational-space.
 * operational-space comes with ABSOLUTELY NO WARRANTY and is licensed under GPL-2.0.
 *    Copyright (C) 2024  VegOwOtenks, Sleppo04
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
#ifndef COMMON_DYNAMICARRAY_DYNAMICARRAY_H_
#define COMMON_DYNAMICARRAY_DYNAMICARRAY_H_

#define DYNAMICARRAY_FOREACH(da, i) for (size_t i = 0; i < (da).reserved; i++)

typedef int (*DynamicArrayFindFunction) (void* element, void* xarg);

// -1 look to the left
// 1 look to the right
// 0 found
typedef DynamicArrayFindFunction DynamicArrayLinearFindFunction;

typedef int (*DynamicArrayCloneFunction) (void* to, void* from, void* xarg);

#include "../pointers/pointers.h"
#include "../allocator-interface/allocator-interface.h"

typedef struct DynamicArray_s {
	size_t object_size;
	size_t capacity;
	size_t reserved;
	void*  memory;

	allocator_t* allocator;
} DynamicArray;


/// @brief Initializes a new dynamic-array at the target memory location
/// @param target target memory location
/// @param object_size size of the objects this array will contain
/// @param initial_capacity expected needed capacity of the array
/// @param allocator allocator that provides memory for this array
/// @return EDESTADDRREQ, ERANGE, EINVAL, ENOMEM, EXIT_SUCCESS
int DynamicArray_Create(DynamicArray* target, size_t object_size, size_t initial_capacity, allocator_t* allocator);

/** @brief Initializes a new dynamic-array at the target, copying data from original
 * This is a 'shallow copy', the elements are by no means deep-copied themselves.
 * Use DeepClone for that
 **/
/// @param target target memory location
/// @param original original dynamic array, data will be copied from here
/// @return EDESTADDRREQ, ERANGE, EINVAL, ENOMEM, EXIT_SUCCESS
int DynamicArray_Clone(DynamicArray* target, DynamicArray* original);

/** @brief Initializes a new dynamic-array at the target, copying data from original
 * This is a 'shallow copy', the elements are by no means deep-copied themselves.
 * Use DeepClone for that
 **/
/// @param target target memory location
/// @param original original dynamic array, data will be copied from here
/// @param clone Function to apply on each element of the array for copying
/// @param xarg Extra Argument that will be supplied to the clone function
/// @return EDESTADDRREQ, ERANGE, EINVAL, ENOMEM, EXIT_SUCCESS
int DynamicArray_DeepClone(DynamicArray* target, DynamicArray* original, DynamicArrayCloneFunction clone, void* xarg);

/// @brief Resize the array to a new capacity, eventually dropping elements out-of-bounds
/// @param array `this`
/// @param new_capacity Capacity the array shall have after resizing
/// @return EDESTADDRREQ, ERANGE, ENOMEM, EXIT_SUCCESS
int DynamicArray_Resize(DynamicArray* array, size_t new_capacity);

/// @brief Resize the objects stored in the array, truncating the values or leaving it uninitialized
/// @param array `this`
/// @param new_object_size New object size to apply
/// @return EDESTADDRREQ, ERANGE, ENOMEM, EXIT_SUCCESS
int DynamicArray_ResizeObjects(DynamicArray* array, size_t new_object_size);

/// @brief Copy object of size array->object_size into the array memory
/// @param array `this`
/// @param object Address of the object that will be copied
/// @return EDESTADDRREQ, ERANGE, ENOMEM, EXIT_SUCCESS
int DynamicArray_Append(DynamicArray* array, void* object);

/// @brief Append a zero-ed structure and return the pointer to it
/// @param array `this`
/// @param pointer The pointer to the structure will be stored here
/// @return EDESTADDRREQ, ERANGE, ENOMEM, EXIT_SUCCESS
int DynamicArray_AppendEmpty(DynamicArray* array, void** pointer);

/// @brief Remove the object at the index from the array, shifting all objects behind to the left
/// @param array `this`
/// @param index Index of the removed object
/// @return EDESTADDRREQ, ERANGE, ENOMEM, EXIT_SUCCESS
int DynamicArray_Remove(DynamicArray* array, size_t index);

/// @brief Remove the object at the index from the array, moving the rightmost object to that index
/// @param array `this`
/// @param index Index to alter
/// @return EDESTADDRREQ, ERANGE, ENOMEM, EXIT_SUCCESS
int DynamicArray_RemoveFast(DynamicArray* array, size_t index);

/// @brief Find a element by applying a function
/// @param array `this`
/// @param function function to call on each element
/// @param xarg extra argument for the called function
/// @return index, SIZE_MAX
size_t DynamicArray_FindFunction(DynamicArray* array, DynamicArrayFindFunction function, void* xarg);

/// @brief Find a element using linear search by applying a function
/// @param array `this`
/// @param function function to call on each element for comparison
/// @param xarg extra argument for the called function, should be some sort of search key
/// @return index, SIZE_MAX
size_t DynamicArray_FindFunctionLinear(DynamicArray* array, DynamicArrayLinearFindFunction function, void* xarg);

/// @brief Return a pointer to the object at the index
/// @param array `this`
/// @param index Retrieved objects index
/// @return NULL, Pointer
void* DynamicArray_GetPointer(DynamicArray* array, size_t index);

/// @brief Calculate the length of the array
/// @param array `this`
/// @return array->reserved
size_t DynamicArray_GetLength(DynamicArray* array);

/// @brief Destroy the array and all contents irreversibly
/// @param array `this`
/// @return void
void DynamicArray_Destroy(DynamicArray* array);

#endif /* SRC_COMMON_DYNAMICARRAY_DYNAMICARRAY_H_ */
